#
# if-nest.R, 12 Oct 24


library("colorspace")
library("plyr")


par(bty="l")
par(las=1)
par(pch="+")
pal_col=rainbow(4)


# line is the number of the line containing the first character of the matched position;
# col is the (0-based) column containing the first character of the matched position;
# line_end is the number of the line containing the last character of the matched position;
# col_end is the (0-based) column containing the last character of the matched position.

# Estimate nesting label based on line number in file
calc_nesting=function(df)
{
# Coccinelle output order is not related to source code order
df=df[order(df$if_line, decreasing=FALSE), ]
df$num=1:nrow(df)

# Create a start/end pair for each if-statement
# +1 for start, -1 for end
seq=data.frame(line_num=c(df$if_line, df$cmpd_e_end),
		val=c(rep(1, nrow(df)), rep(-1, nrow(df))))
seq=seq[order(seq$line_num, decreasing=FALSE), ]

# Nesting is cumulative sum of +/-1s
seq$nesting=cumsum(seq$val)
df$nesting=subset(seq, val == 1)$nesting # remove inserted end points

# Don't double count by including an else arm
return(subset(df, stmt == "if"))
}


process_ifs=function(project)
{
# file;stmt;if_line;if_col;cmpd_end;cmpd_line_end;expr
csv=read.csv(paste0(project, ".csv"), sep=";", quote="")
sim=read.csv(paste0(project, ".sim"), sep=",")

csvsim=cbind(csv, sim)

cs_nest=ddply(csvsim, .(file), calc_nesting)

cs_nest$clauses=1+cs_nest$land+cs_nest$lor

# df=subset(sql, file == "/home/src/sqlite-src-3460100/src/fkey.c")

return(cs_nest)
}

linux=process_ifs("linux-6.11.1")
sql=process_ifs("sqlite")
vim=process_ifs("vim")

nrow(linux)
nrow(sql)
nrow(vim)

lnest_cnt=count(linux$nesting)
snest_cnt=count(sql$nesting)
vnest_cnt=count(vim$nesting)

pal_col=rainbow(3)
plot(lnest_cnt$x, lnest_cnt$freq, log="y", col=pal_col[1],
	xlab="Nesting level", ylab="Occurrences")

points(vnest_cnt$x, vnest_cnt$freq, col=pal_col[2])
points(snest_cnt$x, snest_cnt$freq, col=pal_col[3])

xrange=1:13
vn_mod=glm(log(freq) ~ x, data=vnest_cnt, subset=(x < 12))
summary(vn_mod)

sn_mod=glm(log(freq) ~ x, data=snest_cnt, subset=(x < 12))
summary(sn_mod)

pred=predict(vn_mod, newdata=data.frame(x=xrange))
lines(xrange, exp(pred), col=pal_col[2])
pred=predict(sn_mod, newdata=data.frame(x=xrange))
lines(xrange, exp(pred), col=pal_col[3])

legend("topright", legend=c("Linux", "Vim", "SQLite"), cex=1.2,
                        fill=pal_col, border="white", bty="n")

dev.copy(dev=png, file="LVS-if-nesting.png")
dev.off()


lclause_cnt=count(linux$clauses)
sclause_cnt=count(sql$clauses)
vclause_cnt=count(vim$clauses)

plot(lclause_cnt$x, lclause_cnt$freq, log="xy", col=pal_col[1],
	xlab="Clauses in condition", ylab="Occurrences")
points(vclause_cnt$x, vclause_cnt$freq, col=pal_col[2])
points(sclause_cnt$x, sclause_cnt$freq, col=pal_col[3])

legend("topright", legend=c("Linux", "Vim", "SQLite"), cex=1.2,
                        fill=pal_col, border="white", bty="n")

dev.copy(dev=png, file="LVS-cond-clauses.png")
dev.off()


plot(vim$nesting, vim$clauses,
	xlab="Nesting", ylab="Clauses in condition")

plot(jitter(linux$nesting), jitter(linux$clauses), col="salmon",
	xaxs="i", yaxs="i",
	xlim=c(0, 30), ylim=c(0, 30),
	xlab="Nesting", ylab="Clauses in condition")
dev.copy(dev=png, file="Linux-nest-clause.png")
dev.off()

# Fit linear and non-linear models
linux$inv_nesting=1/linux$nesting
cn_mod=glm(clauses ~ inv_nesting, data=linux, subset=(nesting > 0))
summary(cn_mod)

cn_mod=nls(clauses ~ b/nesting^c, data=linux, subset=(nesting > 0),
		start=list(b=5, c=-1))
pred=predict(cn_mod, newdata=data.frame(nesting=xrange))
lines(xrange, pred, col=pal_col[1])

# verily lots of cpu time
# lines(loess.smooth(linux$nesting, linux$clauses, span=0.7), col=pal_col[1])

# smoothScatter(log(linux$nesting+1e-3), log(linux$clauses),
# 		xlim=c(1, 3), ylim=c(1, 3),
# 		xlab="Nesting", ylab="Clauses")
# 
# smoothScatter(linux$nesting, linux$clauses,
#                 xlim=c(1, 10), ylim=c(1, 10),
#                 xlab="Nesting", ylab="Clauses", nbin=200)

nl_mod=glm(nesting ~ clauses, data=linux, family=poisson)
summary(nl_mod)

